@@ -5,7 +5,7 @@
 
    Copyright (C) 1997-2005 Luke Howard
    Copyright (C) 2007 West Consulting
-   Copyright (C) 2007, 2008, 2009 Arthur de Jong
+   Copyright (C) 2007, 2008, 2009, 2010 Arthur de Jong
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -28,6 +28,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <ctype.h>
 #include <assert.h>
 #include <sys/types.h>
@@ -85,7 +86,7 @@ static void cfg_defaults(struct ldap_config *cfg)
   for (i=0;i<(NSS_LDAP_CONFIG_URI_MAX+1);i++)
   {
     cfg->ldc_uris[i].uri=NULL;
-    cfg->ldc_uris[i].lastok=0;
+    cfg->ldc_uris[i].firstfail=0;
     cfg->ldc_uris[i].lastfail=0;
   }
 #ifdef LDAP_VERSION3
@@ -95,6 +96,7 @@ static void cfg_defaults(struct ldap_config *cfg)
 #endif /* not LDAP_VERSION3 */
   cfg->ldc_binddn=NULL;
   cfg->ldc_bindpw=NULL;
+  cfg->ldc_rootpwmoddn=NULL;
   cfg->ldc_sasl_authcid=NULL;
   cfg->ldc_sasl_authzid=NULL;
   cfg->ldc_sasl_secprops=NULL;
@@ -110,12 +112,14 @@ static void cfg_defaults(struct ldap_config *cfg)
   cfg->ldc_timelimit=LDAP_NO_LIMIT;
   cfg->ldc_idle_timelimit=0;
   cfg->ldc_reconnect_sleeptime=1;
-  cfg->ldc_reconnect_maxsleeptime=10;
+  cfg->ldc_reconnect_retrytime=10;
 #ifdef LDAP_OPT_X_TLS
   cfg->ldc_ssl_on=SSL_OFF;
 #endif /* LDAP_OPT_X_TLS */
   cfg->ldc_restart=1;
   cfg->ldc_pagesize=0;
+  cfg->ldc_nss_initgroups_ignoreusers=NULL;
+  cfg->ldc_pam_authz_search=NULL;
 }
 
 /* simple strdup wrapper */
@@ -161,32 +165,55 @@ static void add_uri(const char *filename,int lnr,
 
 #ifdef HAVE_LDAP_DOMAIN2HOSTLIST
 /* return the domain name of the current host
-   we return part of the structure that is retured by gethostbyname()
-   so there should be no need to free() this entry, however we should
-   use the value before any other call to gethostbyname() */
-static const char *cfg_getdomainname(const char *filename,int lnr)
+   the returned string must be freed by caller */
+static char *cfg_getdomainname(const char *filename,int lnr)
 {
-  char hostname[HOST_NAME_MAX],*domain;
-  struct hostent *host;
-  /* lookup the hostname and with that the fqdn to extract the domain */
+  char hostname[HOST_NAME_MAX+1],*domain;
+  int hostnamelen;
+  int i;
+  struct hostent *host=NULL;
+  /* get system hostname */
   if (gethostname(hostname,sizeof(hostname))<0)
   {
-    log_log(LOG_ERR,"%s:%d: gethostname(): %s",filename,lnr,strerror(errno));
-    exit(EXIT_FAILURE);
+    log_log(LOG_ERR,"%s:%d: gethostname() failed: %s",filename,lnr,strerror(errno));
+    exit (EXIT_FAILURE);
   }
-  if ((host=gethostbyname(hostname))==NULL)
+  hostnamelen=strlen(hostname);
+  /* lookup hostent */
+  host=gethostbyname(hostname);
+  if (host==NULL)
   {
     log_log(LOG_ERR,"%s:%d: gethostbyname(%s): %s",filename,lnr,hostname,hstrerror(h_errno));
     exit(EXIT_FAILURE);
   }
-  /* TODO: this may fail if the fqdn is in h_aliases */
-  if ((domain=strchr(host->h_name,'.'))==NULL)
+  /* check h_name for fqdn starting with our hostname */
+  if ((strncasecmp(hostname,host->h_name,hostnamelen)==0)&&
+      (host->h_name[hostnamelen]=='.')&&
+      (host->h_name[hostnamelen+1]!='\0'))
+    return strdup(host->h_name+hostnamelen+1);
+  /* also check h_aliases */
+  for (i=0;host->h_aliases[i]!=NULL;i++)
   {
-    log_log(LOG_ERR,"%s:%d: host name %s is not in fqdn form",filename,lnr,host->h_name);
-    exit(EXIT_FAILURE);
+    if ((strncasecmp(hostname,host->h_aliases[i],hostnamelen)==0)&&
+        (host->h_aliases[i][hostnamelen]=='.')&&
+        (host->h_aliases[i][hostnamelen+1]!='\0'))
+      return strdup(host->h_aliases[i]+hostnamelen+1);
   }
-  /* we're done */
-  return domain+1;
+  /* fall back to any domain part in h_name */
+  if (((domain=strchr(host->h_name,'.'))!=NULL)&&
+      (domain[1]!='\0'))
+    return strdup(domain+1);
+  /* also check h_aliases */
+  for (i=0;host->h_aliases[i]!=NULL;i++)
+  {
+    if (((domain=strchr(host->h_aliases[i],'.'))!=NULL)&&
+        (domain[1]!='\0'))
+      return strdup(domain+1);
+  }
+  /* we've tried everything now */
+  log_log(LOG_ERR,"%s:%d: unable to determinate a domainname for hostname %s",
+          filename,lnr,hostname);
+  exit(EXIT_FAILURE);
 }
 
 /* add URIs by doing DNS queries for SRV records */
@@ -194,7 +221,7 @@ static void add_uris_from_dns(const char *filename,int lnr,
                               struct ldap_config *cfg)
 {
   int ret=0;
-  const char *domain;
+  char *domain;
   char *hostlist=NULL,*nxt;
   char buf[HOST_NAME_MAX+sizeof("ldap://")];
   domain=cfg_getdomainname(filename,lnr);
@@ -223,6 +250,7 @@ static void add_uris_from_dns(const char *filename,int lnr,
     /* get next entry from list */
     hostlist=nxt;
   }
+  free(domain);
 }
 #endif /* HAVE_LDAP_DOMAIN2HOSTLIST */
 
@@ -540,13 +568,16 @@ static void set_base(const char *filename,int lnr,
                      const char *value,const char **var)
 {
 #ifdef HAVE_LDAP_DOMAIN2DN
+  char *domain = NULL;
   char *domaindn=NULL;
 #endif /* HAVE_LDAP_DOMAIN2DN */
   /* if the base is "DOMAIN" use the domain name */
   if (strcasecmp(value,"domain")==0)
   {
 #ifdef HAVE_LDAP_DOMAIN2DN
-    ldap_domain2dn(cfg_getdomainname(filename,lnr),&domaindn);
+    domain=cfg_getdomainname(filename,lnr);
+    ldap_domain2dn(domain,&domaindn);
+    free(domain);
     log_log(LOG_DEBUG,"set_base(): setting base to %s from domain",domaindn);
     value=domaindn;
 #else /* not HAVE_LDAP_DOMAIN2DN */
@@ -650,6 +681,50 @@ static void parse_map_statement(const char *filename,int lnr,
   }
 }
 
+/* this function modifies the statement argument passed */
+static void parse_nss_initgroups_ignoreusers_statement(
+              const char *filename,int lnr,const char *keyword,
+              char *line,struct ldap_config *cfg)
+{
+  char token[MAX_LINE_LENGTH];
+  char *username,*next;
+  struct passwd *pwent;
+  check_argumentcount(filename,lnr,keyword,(line!=NULL)&&(*line!='\0'));
+  if (cfg->ldc_nss_initgroups_ignoreusers==NULL)
+    cfg->ldc_nss_initgroups_ignoreusers=set_new();
+  while (get_token(&line,token,sizeof(token))!=NULL)
+  {
+    if (strcasecmp(token,"alllocal")==0)
+    {
+      /* go over all users (this will work because nslcd is not yet running) */
+      setpwent();
+      while ((pwent=getpwent())!=NULL)
+        set_add(cfg->ldc_nss_initgroups_ignoreusers,pwent->pw_name);
+      endpwent();
+    }
+    else
+    {
+      next=token;
+      while (*next!='\0')
+      {
+        username=next;
+        /* find the end of the current username */
+        while ((*next!='\0')&&(*next!=',')) next++;
+        if (*next==',')
+        {
+          *next='\0';
+          next++;
+        }
+        /* check if user exists (but add anyway) */
+        pwent=getpwnam(username);
+        if (pwent==NULL)
+          log_log(LOG_ERR,"%s:%d: user '%s' does not exist",filename,lnr,username);
+        set_add(cfg->ldc_nss_initgroups_ignoreusers,username);
+      }
+    }
+  }
+}
+
 static void cfg_read(const char *filename,struct ldap_config *cfg)
 {
   FILE *fp;
@@ -670,7 +745,7 @@ static void cfg_read(const char *filename,struct ldap_config *cfg)
     exit(EXIT_FAILURE);
   }
   /* read file and parse lines */
-  while (fgets(linebuf,MAX_LINE_LENGTH,fp)!=NULL)
+  while (fgets(linebuf,sizeof(linebuf),fp)!=NULL)
   {
     lnr++;
     line=linebuf;
@@ -739,6 +814,10 @@ static void cfg_read(const char *filename,struct ldap_config *cfg)
     {
       get_restdup(filename,lnr,keyword,&line,&cfg->ldc_bindpw);
     }
+    else if (strcasecmp(keyword,"rootpwmoddn")==0)
+    {
+      get_restdup(filename,lnr,keyword,&line,&cfg->ldc_rootpwmoddn);
+    }
     /* SASL authentication options */
     else if (strcasecmp(keyword,"sasl_authcid")==0)
     {
@@ -845,9 +924,12 @@ static void cfg_read(const char *filename,struct ldap_config *cfg)
       get_int(filename,lnr,keyword,&line,&cfg->ldc_reconnect_sleeptime);
       get_eol(filename,lnr,keyword,&line);
     }
-    else if (strcasecmp(keyword,"reconnect_maxsleeptime")==0)
+    else if ( (strcasecmp(keyword,"reconnect_retrytime")==0) ||
+              (strcasecmp(keyword,"reconnect_maxsleeptime")==0) )
     {
-      get_int(filename,lnr,keyword,&line,&cfg->ldc_reconnect_maxsleeptime);
+      if (strcasecmp(keyword,"reconnect_maxsleeptime")==0)
+        log_log(LOG_WARNING,"%s:%d: option %s has been renamed to reconnect_retrytime",filename,lnr,keyword);
+      get_int(filename,lnr,keyword,&line,&cfg->ldc_reconnect_retrytime);
       get_eol(filename,lnr,keyword,&line);
     }
 #ifdef LDAP_OPT_X_TLS
@@ -938,6 +1020,15 @@ static void cfg_read(const char *filename,struct ldap_config *cfg)
       get_int(filename,lnr,keyword,&line,&cfg->ldc_pagesize);
       get_eol(filename,lnr,keyword,&line);
     }
+    else if (strcasecmp(keyword,"nss_initgroups_ignoreusers")==0)
+    {
+      parse_nss_initgroups_ignoreusers_statement(filename,lnr,keyword,line,cfg);
+    }
+    else if (strcasecmp(keyword,"pam_authz_search")==0)
+    {
+      check_argumentcount(filename,lnr,keyword,(line!=NULL)&&(*line!='\0'));
+      cfg->ldc_pam_authz_search=xstrdup(line);
+    }
 #ifdef ENABLE_CONFIGFILE_CHECKING
     /* fallthrough */
     else
@@ -951,6 +1042,59 @@ static void cfg_read(const char *filename,struct ldap_config *cfg)
   fclose(fp);
 }
 
+#ifdef NSLCD_BINDPW_PATH
+static void bindpw_read(const char *filename,struct ldap_config *cfg)
+{
+  FILE *fp;
+  char linebuf[MAX_LINE_LENGTH];
+  int i;
+  /* open config file */
+  errno=0;
+  if ((fp=fopen(filename,"r"))==NULL)
+  {
+    if (errno==ENOENT)
+    {
+      log_log(LOG_DEBUG,"no bindpw file (%s)",filename);
+      return; /* ignore */
+    }
+    else
+    {
+      log_log(LOG_ERR,"cannot open bindpw file (%s): %s",filename,strerror(errno));
+      exit(EXIT_FAILURE);
+    }
+  }
+  /* read the first line */
+  if (fgets(linebuf,sizeof(linebuf),fp)==NULL)
+  {
+    log_log(LOG_ERR,"%s: error reading first line",filename);
+    exit(EXIT_FAILURE);
+  }
+  /* chop the last char off and save the rest as bindpw */
+  i=strlen(linebuf);
+
+  i=(int)strlen(linebuf);
+  if ((i<=0)||(linebuf[i-1]!='\n'))
+  {
+    log_log(LOG_ERR,"%s:1: line too long or missing newline",filename);
+    exit(EXIT_FAILURE);
+  }
+  linebuf[i-1]='\0';
+  if (strlen(linebuf)==0)
+  {
+    log_log(LOG_ERR,"%s:1: the password is empty",filename);
+    exit(EXIT_FAILURE);
+  }
+  cfg->ldc_bindpw=strdup(linebuf);
+  /* check if there is no more data in the file */
+  if (fgets(linebuf,sizeof(linebuf),fp)!=NULL)
+  {
+    log_log(LOG_ERR,"%s:2: there is more than one line in the bindpw file",filename);
+    exit(EXIT_FAILURE);
+  }
+  fclose(fp);
+}
+#endif /* NSLCD_BINDPW_PATH */
+
 /* This function tries to get the LDAP search base from the LDAP server.
    Note that this returns a string that has been allocated with strdup().
    For this to work the myldap module needs enough configuration information
@@ -969,7 +1113,7 @@ static MUST_USE char *get_base_from_rootdse(void)
   session=myldap_create_session();
   assert(session!=NULL);
   /* perform search */
-  search=myldap_search(session,"",LDAP_SCOPE_BASE,"(objectClass=*)",attrs);
+  search=myldap_search(session,"",LDAP_SCOPE_BASE,"(objectClass=*)",attrs,NULL);
   if (search==NULL)
   {
     myldap_session_close(session);
@@ -1022,6 +1166,9 @@ void cfg_init(const char *fname)
   cfg_defaults(nslcd_cfg);
   /* read configfile */
   cfg_read(fname,nslcd_cfg);
+#ifdef NSLCD_BINDPW_PATH
+  bindpw_read(NSLCD_BINDPW_PATH,nslcd_cfg);
+#endif /* NSLCD_BINDPW_PATH */
   /* do some sanity checks */
   if (nslcd_cfg->ldc_uris[0].uri==NULL)
   {
