@@ -5,7 +5,7 @@
 
    Copyright (C) 1997-2005 Luke Howard
    Copyright (C) 2007 West Consulting
-   Copyright (C) 2007, 2008, 2009 Arthur de Jong
+   Copyright (C) 2007, 2008, 2009, 2010 Arthur de Jong
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -32,6 +32,7 @@
 #include <ldap.h>
 
 #include "compat/attrs.h"
+#include "common/set.h"
 
 /* values for uid and gid */
 #define NOUID ((gid_t)-1)
@@ -70,9 +71,9 @@ enum ldap_map_selector
 struct myldap_uri
 {
   char *uri;
-  /* timestamp of last ok search */
-  time_t lastok;
-  /* timestamp of last ok search */
+  /* time of first failed operation */
+  time_t firstfail;
+  /* time of last failed operation */
   time_t lastfail;
 };
 
@@ -92,6 +93,8 @@ struct ldap_config
   char *ldc_binddn;
   /* bind cred */
   char *ldc_bindpw;
+  /* bind DN for password modification by administrator */
+  char *ldc_rootpwmoddn;
   /* sasl authentication id */
   char *ldc_sasl_authcid;
   /* sasl authorization id */
@@ -121,7 +124,7 @@ struct ldap_config
   /* seconds to sleep; doubled until max */
   int ldc_reconnect_sleeptime;
   /* maximum seconds to sleep */
-  int ldc_reconnect_maxsleeptime;
+  int ldc_reconnect_retrytime;
 #ifdef LDAP_OPT_X_TLS
   /* SSL enabled */
   enum ldap_ssl_options ldc_ssl_on;
@@ -130,6 +133,11 @@ struct ldap_config
   int ldc_restart;
   /* set to a greater than 0 to enable handling of paged results with the specified size */
   int ldc_pagesize;
+  /* the users for which no initgroups() searches should be done
+     Note: because we use a set here comparisons will be case-insensitive */
+  SET *ldc_nss_initgroups_ignoreusers;
+  /* the search that should be performed to do autorisation checks */
+  char *ldc_pam_authz_search;
 };
 
 /* this is a pointer to the global configuration, it should be available
