@@ -5,7 +5,7 @@
 
    Copyright (C) 1997-2006 Luke Howard
    Copyright (C) 2006, 2007 West Consulting
-   Copyright (C) 2006, 2007, 2008, 2009 Arthur de Jong
+   Copyright (C) 2006, 2007, 2008, 2009, 2010 Arthur de Jong
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -363,7 +363,9 @@ static int do_sasl_interact(LDAP UNUSED(*ld),unsigned UNUSED(flags),void *defaul
    This returns an LDAP result code. */
 static int do_bind(MYLDAP_SESSION *session,const char *uri)
 {
+#ifdef LDAP_OPT_X_TLS
   int rc;
+#endif /* LDAP_OPT_X_TLS */
 #ifdef HAVE_LDAP_SASL_INTERACTIVE_BIND_S
 #ifndef HAVE_SASL_INTERACT_T
   struct berval cred;
@@ -378,8 +380,9 @@ static int do_bind(MYLDAP_SESSION *session,const char *uri)
     rc=ldap_start_tls_s(session->ld,NULL,NULL);
     if (rc!=LDAP_SUCCESS)
     {
-      log_log(LOG_WARNING,"ldap_start_tls_s() failed: %s: %s (uri=\"%s\")",
-                          ldap_err2string(rc),strerror(errno),uri);
+      log_log(LOG_WARNING,"ldap_start_tls_s() failed: %s%s%s (uri=\"%s\")",
+                          ldap_err2string(rc),(errno==0)?"":": ",
+                          (errno==0)?"":strerror(errno),uri);
       return rc;
     }
   }
@@ -389,7 +392,7 @@ static int do_bind(MYLDAP_SESSION *session,const char *uri)
   {
     /* do a simple bind */
     log_log(LOG_DEBUG,"ldap_simple_bind_s(\"%s\",%s) (uri=\"%s\")",session->binddn,
-                      (session->bindpw[0]!='\0')?"\"*****\"":"empty",uri);
+                      (session->bindpw[0]!='\0')?"\"***\"":"\"\"",uri);
     return ldap_simple_bind_s(session->ld,session->binddn,session->bindpw);
   }
 #ifdef HAVE_LDAP_SASL_INTERACTIVE_BIND_S
@@ -400,10 +403,10 @@ static int do_bind(MYLDAP_SESSION *session,const char *uri)
     /* do a simple bind */
     if (nslcd_cfg->ldc_binddn)
       log_log(LOG_DEBUG,"ldap_simple_bind_s(\"%s\",%s) (uri=\"%s\")",nslcd_cfg->ldc_binddn,
-                        nslcd_cfg->ldc_bindpw?"\"*****\"":"NULL",uri);
+                        nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
     else
       log_log(LOG_DEBUG,"ldap_simple_bind_s(NULL,%s) (uri=\"%s\")",
-                        nslcd_cfg->ldc_bindpw?"\"*****\"":"NULL",uri);
+                        nslcd_cfg->ldc_bindpw?"\"***\"":"NULL",uri);
     return ldap_simple_bind_s(session->ld,nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_bindpw);
 #ifdef HAVE_LDAP_SASL_INTERACTIVE_BIND_S
   }
@@ -433,7 +436,7 @@ static int do_bind(MYLDAP_SESSION *session,const char *uri)
       cred.bv_len=0;
     }
     log_log(LOG_DEBUG,"ldap_sasl_bind_s(\"%s\",\"%s\",%s)",nslcd_cfg->ldc_binddn,
-                      nslcd_cfg->ldc_sasl_mech,nslcd_cfg->ldc_bindpw?"\"*****\"":"NULL");
+                      nslcd_cfg->ldc_sasl_mech,nslcd_cfg->ldc_bindpw?"\"***\"":"NULL");
     return ldap_sasl_bind_s(session->ld,nslcd_cfg->ldc_binddn,nslcd_cfg->ldc_sasl_mech,&cred,NULL,NULL,NULL);
 #endif /* not HAVE_SASL_INTERACT_T */
   }
@@ -460,7 +463,9 @@ static int do_set_options(MYLDAP_SESSION *session)
   /* FIXME: move this to a global initialisation routine */
   int rc;
   struct timeval tv;
+#ifdef LDAP_OPT_X_TLS
   int i;
+#endif /* LDAP_OPT_X_TLS */
 #ifdef HAVE_LDAP_SET_REBIND_PROC
   /* the rebind function that is called when chasing referrals, see
      http://publib.boulder.ibm.com/infocenter/iseries/v5r3/topic/apis/ldap_set_rebind_proc.htm
@@ -569,7 +574,6 @@ static void do_close(MYLDAP_SESSION *session)
 static void myldap_session_check(MYLDAP_SESSION *session)
 {
   int i;
-  int runningsearches=0;
   time_t current_time;
   /* check parameters */
   if (!is_valid_session(session))
@@ -581,24 +585,16 @@ static void myldap_session_check(MYLDAP_SESSION *session)
   /* check if we should time out the connection */
   if ((session->ld!=NULL)&&(nslcd_cfg->ldc_idle_timelimit>0))
   {
-    /* check if we have any running searches */
+    /* if we have any running searches, don't time out */
     for (i=0;i<MAX_SEARCHES_IN_SESSION;i++)
-    {
       if ((session->searches[i]!=NULL)&&(session->searches[i]->valid))
-      {
-        runningsearches=1;
-        break;
-      }
-    }
-    /* only consider timeout if we have no running searches */
-    if (!runningsearches)
+        return;
+    /* consider timeout (there are no running searches) */
+    time(&current_time);
+    if ((session->lastactivity+nslcd_cfg->ldc_idle_timelimit)<current_time)
     {
-      time(&current_time);
-      if ((session->lastactivity+nslcd_cfg->ldc_idle_timelimit)<current_time)
-      {
-        log_log(LOG_DEBUG,"do_open(): idle_timelimit reached");
-        do_close(session);
-      }
+      log_log(LOG_DEBUG,"myldap_session_check(): idle_timelimit reached");
+      do_close(session);
     }
   }
 }
@@ -625,9 +621,10 @@ static int do_open(MYLDAP_SESSION *session)
   rc=ldap_initialize(&(session->ld),nslcd_cfg->ldc_uris[session->current_uri].uri);
   if (rc!=LDAP_SUCCESS)
   {
-    log_log(LOG_WARNING,"ldap_initialize(%s) failed: %s: %s",
+    log_log(LOG_WARNING,"ldap_initialize(%s) failed: %s%s%s",
                         nslcd_cfg->ldc_uris[session->current_uri].uri,
-                        ldap_err2string(rc),strerror(errno));
+                        ldap_err2string(rc),(errno==0)?"":": ",
+                        (errno==0)?"":strerror(errno));
     if (session->ld!=NULL)
     {
       log_log(LOG_DEBUG,"ldap_unbind()");
@@ -659,9 +656,11 @@ static int do_open(MYLDAP_SESSION *session)
   if (rc!=LDAP_SUCCESS)
   {
     /* log actual LDAP error code */
-    log_log(LOG_WARNING,"failed to bind to LDAP server %s: %s: %s",
+    log_log((session->binddn[0]=='\0')?LOG_WARNING:LOG_DEBUG,
+                        "failed to bind to LDAP server %s: %s%s%s",
                         nslcd_cfg->ldc_uris[session->current_uri].uri,
-                        ldap_err2string(rc),strerror(errno));
+                        ldap_err2string(rc),(errno==0)?"":": ",
+                        (errno==0)?"":strerror(errno));
     rc2=ldap_unbind(session->ld);
     session->ld=NULL;
     if (rc2!=LDAP_SUCCESS)
@@ -676,13 +675,11 @@ static int do_open(MYLDAP_SESSION *session)
   }
   /* update last activity and finish off state */
   time(&(session->lastactivity));
-  log_log(LOG_INFO,"connected to LDAP server %s",
-                   nslcd_cfg->ldc_uris[session->current_uri].uri);
   return LDAP_SUCCESS;
 }
 
 /* Set alternative credentials for the session. */
-int myldap_set_credentials(MYLDAP_SESSION *session,const char *dn,
+void myldap_set_credentials(MYLDAP_SESSION *session,const char *dn,
                            const char *password)
 {
   /* copy dn and password into session */
@@ -690,8 +687,6 @@ int myldap_set_credentials(MYLDAP_SESSION *session,const char *dn,
   session->binddn[sizeof(session->binddn)-1]='\0';
   strncpy(session->bindpw,password,sizeof(session->bindpw));
   session->bindpw[sizeof(session->bindpw)-1]='\0';
-  /* try to open a connection */
-  return do_open(session);
 }
 
 static int do_try_search(MYLDAP_SEARCH *search)
@@ -802,30 +797,48 @@ static int do_retry_search(MYLDAP_SEARCH *search)
   time_t nexttry;
   time_t t;
   int rc=LDAP_UNAVAILABLE;
+  struct myldap_uri *current_uri;
+  int dotry[NSS_LDAP_CONFIG_URI_MAX];
+  /* clear time stamps */
+  for (start_uri=0;start_uri<NSS_LDAP_CONFIG_URI_MAX;start_uri++)
+    dotry[start_uri]=1;
   /* keep trying until we time out */
-  endtime=time(NULL)+nslcd_cfg->ldc_reconnect_maxsleeptime;
-  nexttry=endtime;
+  endtime=time(NULL)+nslcd_cfg->ldc_reconnect_retrytime;
   while (1)
   {
+    nexttry=endtime;
     /* try each configured URL once */
     pthread_mutex_lock(&uris_mutex);
     start_uri=search->session->current_uri;
     do
     {
-      /* only try if we haven't just had an error and it was a long tme
-         since the last ok */
-      if ( ( ( nslcd_cfg->ldc_uris[search->session->current_uri].lastfail -
-               nslcd_cfg->ldc_uris[search->session->current_uri].lastok ) < nslcd_cfg->ldc_reconnect_maxsleeptime) ||
-           ( time(NULL) >= (nslcd_cfg->ldc_uris[search->session->current_uri].lastfail+nslcd_cfg->ldc_reconnect_maxsleeptime) ) )
+      current_uri=&(nslcd_cfg->ldc_uris[search->session->current_uri]);
+      /* only try this URI if we should */
+      if (!dotry[search->session->current_uri])
+      { /* skip this URI */ }
+      else if ( (current_uri->lastfail > (current_uri->firstfail+nslcd_cfg->ldc_reconnect_retrytime)) &&
+                ((t=time(NULL)) < (current_uri->lastfail+nslcd_cfg->ldc_reconnect_retrytime)) )
+      {
+        /* we are in a hard fail state and have retried not long ago */
+        log_log(LOG_DEBUG,"not retrying server %s which failed just %d second(s) ago and has been failing for %d seconds",
+                          current_uri->uri,(int)(t-current_uri->lastfail),
+                          (int)(t-current_uri->firstfail));
+        dotry[search->session->current_uri]=0;
+      }
+      else
       {
-        pthread_mutex_unlock(&uris_mutex);
         /* try to start the search */
+        pthread_mutex_unlock(&uris_mutex);
         rc=do_try_search(search);
         if (rc==LDAP_SUCCESS)
         {
-          /* update ok time and return search handle */
+          /* update ok time */
           pthread_mutex_lock(&uris_mutex);
-          nslcd_cfg->ldc_uris[search->session->current_uri].lastok=time(NULL);
+          current_uri->firstfail=0;
+          current_uri->lastfail=0;
+          /* check if we are coming back from an error */
+          if ((current_uri->lastfail>0)||(search->session->current_uri!=start_uri))
+            log_log(LOG_INFO,"connected to LDAP server %s",current_uri->uri);
           pthread_mutex_unlock(&uris_mutex);
           /* flag the search as valid */
           search->valid=1;
@@ -836,17 +849,24 @@ static int do_retry_search(MYLDAP_SEARCH *search)
         /* update time of failure and figure out when we should retry */
         pthread_mutex_lock(&uris_mutex);
         t=time(NULL);
-        nslcd_cfg->ldc_uris[search->session->current_uri].lastfail=t;
-        t+=nslcd_cfg->ldc_reconnect_sleeptime;
-        if (t<nexttry)
-          nexttry=t;
-      }
-      else if (nslcd_cfg->ldc_uris[search->session->current_uri].lastfail>0)
-      {
-        /* we are in a hard fail state, figure out when we can retry */
-        t=(nslcd_cfg->ldc_uris[search->session->current_uri].lastfail+nslcd_cfg->ldc_reconnect_maxsleeptime);
-        if (t<nexttry)
-          nexttry=t;
+        /* update timestaps unless we are doing an authentication search */
+        if (search->session->binddn[0]=='\0')
+        {
+          if (current_uri->firstfail==0)
+            current_uri->firstfail=t;
+          current_uri->lastfail=t;
+        }
+        /* if it is one of these, retrying this URI is not going to help */
+        if ((rc==LDAP_INVALID_CREDENTIALS)||(rc==LDAP_INSUFFICIENT_ACCESS)||
+            (rc==LDAP_AUTH_METHOD_NOT_SUPPORTED))
+          dotry[search->session->current_uri]=0;
+        /* check whether we should try this URI again */
+        else if (t <= (current_uri->firstfail+nslcd_cfg->ldc_reconnect_retrytime))
+        {
+          t+=nslcd_cfg->ldc_reconnect_sleeptime;
+          if (t<nexttry)
+            nexttry=t;
+        }
       }
       /* try the next URI (with wrap-around) */
       search->session->current_uri++;
@@ -858,33 +878,35 @@ static int do_retry_search(MYLDAP_SEARCH *search)
     /* see if it is any use sleeping */
     if (nexttry>=endtime)
     {
-      log_log(LOG_ERR,"no available LDAP server found");
+      if (search->session->binddn[0]=='\0')
+        log_log(LOG_ERR,"no available LDAP server found");
       return rc;
     }
     /* sleep between tries */
     sleeptime=nexttry-time(NULL);
-    if (sleeptime>nslcd_cfg->ldc_reconnect_maxsleeptime)
-      sleeptime=nslcd_cfg->ldc_reconnect_maxsleeptime;
     if (sleeptime>0)
     {
       log_log(LOG_WARNING,"no available LDAP server found, sleeping %d seconds",sleeptime);
       (void)sleep(sleeptime);
     }
-    nexttry=time(NULL)+nslcd_cfg->ldc_reconnect_maxsleeptime;
   }
 }
 
 MYLDAP_SEARCH *myldap_search(
         MYLDAP_SESSION *session,
-        const char *base,int scope,const char *filter,const char **attrs)
+        const char *base,int scope,const char *filter,const char **attrs,
+        int *rcp)
 {
   MYLDAP_SEARCH *search;
   int i;
+  int rc;
   /* check parameters */
   if (!is_valid_session(session)||(base==NULL)||(filter==NULL)||(attrs==NULL))
   {
     log_log(LOG_ERR,"myldap_search(): invalid parameter passed");
     errno=EINVAL;
+    if (rcp!=NULL)
+      *rcp=LDAP_OPERATIONS_ERROR;
     return NULL;
   }
   /* log the call */
@@ -900,16 +922,23 @@ MYLDAP_SEARCH *myldap_search(
     log_log(LOG_ERR,"myldap_search(): too many searches registered with session (max %d)",
                     MAX_SEARCHES_IN_SESSION);
     myldap_search_close(search);
+    if (rcp!=NULL)
+      *rcp=LDAP_OPERATIONS_ERROR;
     return NULL;
   }
   /* regsiter search with the session so we can free it later on */
   session->searches[i]=search;
   /* do the search with retries to all configured servers */
-  if (do_retry_search(search)!=LDAP_SUCCESS)
+  rc=do_retry_search(search);
+  if (rc!=LDAP_SUCCESS)
   {
     myldap_search_close(search);
+    if (rcp!=NULL)
+      *rcp=rc;
     return NULL;
   }
+  if (rcp!=NULL)
+    *rcp=LDAP_SUCCESS;
   return search;
 }
 
@@ -1287,7 +1316,7 @@ static SET *myldap_get_ranged_values(MYLDAP_ENTRY *entry,const char *attr)
     if (search!=NULL)
       myldap_search_close(search);
     /* start the new search */
-    search=myldap_search(session,dn,LDAP_SCOPE_BASE,"(objectClass=*)",attrs);
+    search=myldap_search(session,dn,LDAP_SCOPE_BASE,"(objectClass=*)",attrs,NULL);
     if (search==NULL)
       break;
     entry=myldap_get_entry(search,NULL);
@@ -1520,7 +1549,7 @@ int myldap_escape(const char *src,char *buffer,size_t buflen)
   for (;*src!='\0';src++)
   {
     /* check if char will fit */
-    if (pos>=(buflen+4))
+    if ((pos+4)>=buflen)
       return -1;
     /* do escaping for some characters */
     switch (*src)
@@ -1602,32 +1631,45 @@ int myldap_passwd(
   int rc;
   struct berval ber_userdn, ber_oldpassword, ber_newpassword, ber_retpassword;
   /* check parameters */
-  if (!is_valid_session(session)||(userdn==NULL)||(oldpassword==NULL)||(newpasswd==NULL))
+  if (!is_valid_session(session)||(userdn==NULL)||(newpasswd==NULL))
   {
-    log_log(LOG_ERR,"myldap_exop_passwd(): invalid parameter passed");
+    log_log(LOG_ERR,"myldap_passwd(): invalid parameter passed");
     errno=EINVAL;
     return LDAP_OTHER;
   }
   /* log the call */
-  log_log(LOG_DEBUG,"myldap_exop_passwd(userdn=\"%s\")",userdn);
+  log_log(LOG_DEBUG,"myldap_passwd(userdn=\"%s\",oldpasswd=%s,newpasswd=\"***\")",
+                    userdn,oldpassword?"\"***\"":"NULL");
   /* translate to ber stuff */
-  ber_userdn.bv_val=userdn;
+  ber_userdn.bv_val=(char *)userdn;
   ber_userdn.bv_len=strlen(userdn);
-  ber_oldpassword.bv_val=oldpassword;
-  ber_oldpassword.bv_len=strlen(oldpassword);
-  ber_newpassword.bv_val=newpasswd;
+  ber_newpassword.bv_val=(char *)newpasswd;
   ber_newpassword.bv_len=strlen(newpasswd);
   ber_retpassword.bv_val=NULL;
   ber_retpassword.bv_len=0;
   /* perform request */
-  rc=ldap_passwd_s(session->ld,&ber_userdn,&ber_oldpassword,&ber_newpassword,
-                   &ber_retpassword,NULL,NULL);
+  log_log(LOG_DEBUG,"myldap_passwd(): try ldap_passwd_s() without old password");
+  rc=ldap_passwd_s(session->ld,&ber_userdn,NULL,
+                   &ber_newpassword,&ber_retpassword,NULL,NULL);
   if (rc!=LDAP_SUCCESS)
-    log_log(LOG_ERR,"ldap_passwd_s() failed: %s",ldap_err2string(rc));
-
-
-  /* FIXME: free ber_retpassword data if bv_val!=NULL */
-
+    log_log(LOG_ERR,"ldap_passwd_s() without old password failed: %s",ldap_err2string(rc));
+  /* free returned data if needed */
+  if (ber_retpassword.bv_val!=NULL)
+    ldap_memfree(ber_retpassword.bv_val);
+  if ((rc!=LDAP_SUCCESS)&&(oldpassword!=NULL))
+  {
+    /* retry with old password */
+    log_log(LOG_DEBUG,"myldap_passwd(): try ldap_passwd_s() with old password");
+    ber_oldpassword.bv_val=(char *)oldpassword;
+    ber_oldpassword.bv_len=strlen(oldpassword);
+    /* perform request */
+    rc=ldap_passwd_s(session->ld,&ber_userdn,&ber_oldpassword,
+                     &ber_newpassword,&ber_retpassword,NULL,NULL);
+    if (rc!=LDAP_SUCCESS)
+      log_log(LOG_ERR,"ldap_passwd_s() with old password failed: %s",ldap_err2string(rc));
+    /* free returned data if needed */
+    if (ber_retpassword.bv_val!=NULL)
+      ldap_memfree(ber_retpassword.bv_val);
+  }
   return rc;
-
 }
