@@ -2,7 +2,7 @@
    pam.c - pam processing routines
 
    Copyright (C) 2009 Howard Chu
-   Copyright (C) 2009 Arthur de Jong
+   Copyright (C) 2009, 2010 Arthur de Jong
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -28,12 +28,19 @@
 #ifdef HAVE_STDINT_H
 #include <stdint.h>
 #endif /* HAVE_STDINT_H */
+#include <unistd.h>
 
 #include "common.h"
 #include "log.h"
 #include "myldap.h"
 #include "cfg.h"
 #include "attmap.h"
+#include "common/dict.h"
+#include "common/expr.h"
+
+#ifndef HOST_NAME_MAX
+#define HOST_NAME_MAX 255
+#endif /* not HOST_NAME_MAX */
 
 /* set up a connection and try to bind with the specified DN and password
    returns a NSLCD_PAM_* error code */
@@ -47,15 +54,11 @@ static int try_bind(const char *userdn,const char *password)
   if (session==NULL)
     return NSLCD_PAM_AUTH_ERR;
   /* set up credentials for the session */
-  rc=myldap_set_credentials(session,userdn,password);
-  /* TODO: test rc */
-  if (rc==LDAP_SUCCESS)
-  {
-    /* perform search for own object (just to do any kind of search) */
-    username=lookup_dn2uid(session,userdn,&rc);
-    if (username!=NULL)
-      free(username);
-  }
+  myldap_set_credentials(session,userdn,password);
+  /* perform search for own object (just to do any kind of search) */
+  username=lookup_dn2uid(session,userdn,&rc);
+  if (username!=NULL)
+    free(username);
   /* close the session */
   myldap_session_close(session);
   /* handle the results */
@@ -145,8 +148,18 @@ int nslcd_pam_authc(TFILE *fp,MYLDAP_SESSION *session)
   /* write the response header */
   WRITE_INT32(fp,NSLCD_VERSION);
   WRITE_INT32(fp,NSLCD_ACTION_PAM_AUTHC);
-  /* validate request and fill in the blanks */
-  if (validate_user(session,userdn,sizeof(userdn),username,sizeof(username)))
+  /* if the username is blank and rootpwmoddn is configure, try to authenticate
+     as administrator, otherwise validate request as usual */
+  if ((*username=='\0')&&(nslcd_cfg->ldc_rootpwmoddn!=NULL))
+  {
+    if (strlen(nslcd_cfg->ldc_rootpwmoddn)>=sizeof(userdn))
+    {
+      log_log(LOG_ERR,"nslcd_pam_authc(): rootpwmoddn will not fit in userdn");
+      return -1;
+    }
+    strcpy(userdn,nslcd_cfg->ldc_rootpwmoddn);
+  }
+  else if (validate_user(session,userdn,sizeof(userdn),username,sizeof(username)))
   {
     WRITE_INT32(fp,NSLCD_RESULT_END);
     return -1;
@@ -164,6 +177,94 @@ int nslcd_pam_authc(TFILE *fp,MYLDAP_SESSION *session)
   return 0;
 }
 
+static void autzsearch_var_add(DICT *dict,const char *name,const char *value)
+{
+  size_t sz;
+  char *escaped_value;
+  /* allocate memory for escaped string */
+  sz=((strlen(value)+8)*120)/100;
+  escaped_value=(char *)malloc(sz);
+  if (escaped_value==NULL)
+  {
+    log_log(LOG_CRIT,"autzsearch_var_add(): malloc() failed to allocate memory");
+    return;
+  }
+  /* perform escaping of the value */
+  if(myldap_escape(value,escaped_value,sz))
+  {
+    log_log(LOG_CRIT,"autzsearch_var_add(): myldap_escape() failed to fit in buffer");
+    return;
+  }
+  /* add to dict */
+  dict_put(dict,name,escaped_value);
+}
+
+static void autzsearch_vars_free(DICT *dict)
+{
+  int i;
+  const char **keys;
+  void *value;
+  /* go over all keys and free all the values
+     (they were allocated in autzsearch_var_add) */
+  /* loop over dictionary contents */
+  keys=dict_keys(dict);
+  for (i=0;keys[i]!=NULL;i++)
+  {
+    value=dict_get(dict,keys[i]);
+    if (value)
+      free(value);
+  }
+  free(keys);
+  /* after this values from the dict should obviously no longer be used */
+}
+
+static const char *autzsearch_var_get(const char *name,void *expander_attr)
+{
+  DICT *dict=(DICT *)expander_attr;
+  return (const char *)dict_get(dict,name);
+  /* TODO: if not set use entry to get attribute name (entry can be an
+           element in the dict) */
+}
+
+static int try_autzsearch(MYLDAP_SESSION *session,DICT *dict,const char *searchfilter)
+{
+  char filter_buffer[1024];
+  MYLDAP_SEARCH *search;
+  MYLDAP_ENTRY *entry;
+  static const char *attrs[2];
+  int rc;
+  /* build the search filter */
+  if (expr_parse(searchfilter,filter_buffer,sizeof(filter_buffer),
+                 autzsearch_var_get,(void *)dict)==NULL)
+  {
+    log_log(LOG_ERR,"pam_authz_search \"%s\" is invalid",searchfilter);
+    return -1;
+  }
+  log_log(LOG_DEBUG,"trying pam_authz_search \"%s\"",filter_buffer);
+  /* perform the search */
+  attrs[0]="dn";
+  attrs[1]=NULL;
+  /* FIXME: this only searches the first base */
+  search=myldap_search(session,nslcd_cfg->ldc_bases[0],LDAP_SCOPE_SUBTREE,
+                       filter_buffer,attrs,&rc);
+  if (search==NULL)
+  {
+    log_log(LOG_ERR,"pam_authz_search \"%s\" failed: %s",
+            filter_buffer,ldap_err2string(rc));
+    return -1;
+  }
+  /* try to get an entry */
+  entry=myldap_get_entry(search,NULL);
+  if (entry==NULL)
+  {
+    log_log(LOG_ERR,"pam_authz_search \"%s\" found no matches",filter_buffer);
+    return -1;
+  }
+  log_log(LOG_DEBUG,"pam_authz_search found \"%s\"",myldap_get_dn(entry));
+  /* we've found an entry so it's OK */
+  return 0;
+}
+
 /* check authorisation of the user */
 int nslcd_pam_authz(TFILE *fp,MYLDAP_SESSION *session)
 {
@@ -174,6 +275,8 @@ int nslcd_pam_authz(TFILE *fp,MYLDAP_SESSION *session)
   char ruser[32];
   char rhost[256];
   char tty[256];
+  char hostname[HOST_NAME_MAX+1];
+  DICT *dict;
   /* read request parameters */
   READ_STRING(fp,username);
   READ_STRING(fp,userdn);
@@ -193,7 +296,32 @@ int nslcd_pam_authz(TFILE *fp,MYLDAP_SESSION *session)
     WRITE_INT32(fp,NSLCD_RESULT_END);
     return -1;
   }
-  /* TODO: perform any authorisation checks */
+  if (nslcd_cfg->ldc_pam_authz_search)
+  {
+    /* TODO: perform any authorisation checks */
+    dict=dict_new();
+    autzsearch_var_add(dict,"username",username);
+    autzsearch_var_add(dict,"service",servicename);
+    autzsearch_var_add(dict,"ruser",ruser);
+    autzsearch_var_add(dict,"rhost",rhost);
+    autzsearch_var_add(dict,"tty",tty);
+    if (gethostname(hostname,sizeof(hostname))==0)
+      autzsearch_var_add(dict,"hostname",hostname);
+    /* TODO: fqdn */
+    autzsearch_var_add(dict,"dn",userdn);
+    autzsearch_var_add(dict,"uid",username);
+    if (try_autzsearch(session,dict,nslcd_cfg->ldc_pam_authz_search))
+    {
+      WRITE_INT32(fp,NSLCD_RESULT_BEGIN);
+      WRITE_STRING(fp,username);
+      WRITE_STRING(fp,userdn);
+      WRITE_INT32(fp,NSLCD_PAM_PERM_DENIED);  /* authz */
+      WRITE_STRING(fp,"LDAP authorisation check failed"); /* authzmsg */
+      WRITE_INT32(fp,NSLCD_RESULT_END);
+    }
+    autzsearch_vars_free(dict);
+    dict_free(dict);
+  }
   /* write response */
   WRITE_INT32(fp,NSLCD_RESULT_BEGIN);
   WRITE_STRING(fp,username);
@@ -262,19 +390,28 @@ int nslcd_pam_sess_c(TFILE *fp,MYLDAP_SESSION *session)
   return 0;
 }
 
-static int try_pwmod(const char *userdn,const char *oldpassword,
-                     const char *newpassword)
+static int try_pwmod(const char *binddn,const char *userdn,
+                     const char *oldpassword,const char *newpassword)
 {
   MYLDAP_SESSION *session;
+  char *username;
   int rc;
   /* set up a new connection */
   session=myldap_create_session();
   if (session==NULL)
     return NSLCD_PAM_AUTH_ERR;
   /* set up credentials for the session */
-  rc=myldap_set_credentials(session,userdn,oldpassword);
+  myldap_set_credentials(session,binddn,oldpassword);
+  /* perform search for own object (just to do any kind of search) */
+  username=lookup_dn2uid(session,userdn,&rc);
+  if (username!=NULL)
+    free(username);
+  /* perform actual password modification */
   if (rc==LDAP_SUCCESS)
   {
+    /* if doing password modification as admin, don't pass old password along */
+    if ((nslcd_cfg->ldc_rootpwmoddn!=NULL)&&(strcmp(binddn,nslcd_cfg->ldc_rootpwmoddn)==0))
+      oldpassword=NULL;
     /* perform password modification */
     rc=myldap_passwd(session,userdn,oldpassword,newpassword);
   }
@@ -292,6 +429,7 @@ int nslcd_pam_pwmod(TFILE *fp,MYLDAP_SESSION *session)
   char servicename[64];
   char oldpassword[64];
   char newpassword[64];
+  char *binddn=userdn; /* the user performing the modification */
   int rc;
   /* read request parameters */
   READ_STRING(fp,username);
@@ -306,6 +444,12 @@ int nslcd_pam_pwmod(TFILE *fp,MYLDAP_SESSION *session)
   /* write the response header */
   WRITE_INT32(fp,NSLCD_VERSION);
   WRITE_INT32(fp,NSLCD_ACTION_PAM_PWMOD);
+  /* check if the the user passed the rootpwmoddn */
+  if ((nslcd_cfg->ldc_rootpwmoddn!=NULL)&&(strcmp(userdn,nslcd_cfg->ldc_rootpwmoddn)==0))
+  {
+    binddn=nslcd_cfg->ldc_rootpwmoddn;
+    userdn[0]='\0'; /* cause validate_user() to get the user DN */
+  }
   /* validate request and fill in the blanks */
   if (validate_user(session,userdn,sizeof(userdn),username,sizeof(username)))
   {
@@ -313,7 +457,7 @@ int nslcd_pam_pwmod(TFILE *fp,MYLDAP_SESSION *session)
     return -1;
   }
   /* perform password modification */
-  rc=try_pwmod(userdn,oldpassword,newpassword);
+  rc=try_pwmod(binddn,userdn,oldpassword,newpassword);
   /* write response */
   WRITE_INT32(fp,NSLCD_RESULT_BEGIN);
   WRITE_STRING(fp,username);
