@@ -2,7 +2,7 @@
    pam.c - pam module functions
 
    Copyright (C) 2009 Howard Chu
-   Copyright (C) 2009 Arthur de Jong
+   Copyright (C) 2009, 2010 Arthur de Jong
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -28,10 +28,7 @@
 #include <syslog.h>
 #include <unistd.h>
 #include <sys/types.h>
-
-#include "common.h"
-#include "compat/attrs.h"
-#include "compat/pam_compat.h"
+#include <pwd.h>
 
 /* these are defined (before including pam_modules.h) for staticly linking */
 #define PAM_SM_AUTH
@@ -39,6 +36,10 @@
 #define PAM_SM_SESSION
 #define PAM_SM_PASSWORD
 
+#include "common.h"
+#include "compat/attrs.h"
+#include "compat/pam_compat.h"
+
 #ifdef HAVE_SECURITY_PAM_APPL_H
 #include <security/pam_appl.h>
 #endif /* HAVE_SECURITY_PAM_APPL_H */
@@ -51,23 +52,13 @@
 #include <pam/pam_modules.h>
 #endif /* not HAVE_PAM_PAM_MODULES_H */
 
-#define IGNORE_UNKNOWN  1
-#define IGNORE_UNAVAIL  2
-
-#define USE_FIRST 1
-#define TRY_FIRST 2
-#define USE_TOKEN 4
-
+/* the name we store our context under */
 #define PLD_CTX "PAM_LDAPD_CTX"
 
-/* some systems don't have LOG_AUTHPRIV */
-#ifndef LOG_AUTHPRIV
-#define LOG_AUTHPRIV LOG_AUTH
-#endif /* not LOG_AUTHPRIV */
 
-/* this struct represents that context that the PAM module keeps
+/* this struct represents the context that the PAM module keeps
    between calls */
-typedef struct pld_ctx {
+struct pld_ctx {
   char *user;
   char *dn;
   char *tmpluser;
@@ -77,7 +68,161 @@ typedef struct pld_ctx {
   int authz;
   int sessid;
   char buf[1024];
-} pld_ctx;
+};
+
+/* clear the context to all empty values */
+static void ctx_clear(struct pld_ctx *ctx)
+{
+  if (ctx->user)
+  {
+    free(ctx->user);
+    ctx->user=NULL;
+  }
+  if (ctx->oldpassword)
+  {
+    memset(ctx->oldpassword,0,strlen(ctx->oldpassword));
+    free(ctx->oldpassword);
+    ctx->oldpassword=NULL;
+  }
+  ctx->dn=NULL;
+  ctx->tmpluser=NULL;
+  ctx->authzmsg=NULL;
+  ctx->authok=0;
+  ctx->authz=0;
+}
+
+/* free the context (this is installed as handler into PAM) */
+static void ctx_free(pam_handle_t *UNUSED(pamh),void *data,int UNUSED(err))
+{
+  struct pld_ctx *ctx=data;
+  ctx_clear(ctx);
+  free(ctx);
+}
+
+/* try to get the module's context, returns a PAM status code */
+static int ctx_get(pam_handle_t *pamh,const char *username,struct pld_ctx **pctx)
+{
+  struct pld_ctx *ctx=NULL;
+  int rc;
+  /* try to get the context from PAM */
+  rc=pam_get_data(pamh,PLD_CTX,(const void **)&ctx);
+  if ((rc==PAM_SUCCESS)&&(ctx!=NULL))
+  {
+    /* if the user is different clear the context */
+    if ((ctx->user!=NULL)&&(strcmp(ctx->user,username)!=0))
+      ctx_clear(ctx);
+  }
+  else
+  {
+    /* allocate a new context */
+    ctx=calloc(1,sizeof(struct pld_ctx));
+    if (ctx==NULL)
+    {
+      pam_syslog(pamh,LOG_CRIT,"calloc(): failed to allocate memory: %s",strerror(errno));
+      return PAM_BUF_ERR;
+    }
+    ctx_clear(ctx);
+    /* store the new context with the handler to free it */
+    rc=pam_set_data(pamh,PLD_CTX,ctx,ctx_free);
+    if (rc!=PAM_SUCCESS)
+    {
+      ctx_free(pamh,ctx,0);
+      pam_syslog(pamh,LOG_ERR,"failed to store context: %s",pam_strerror(pamh,rc));
+      return rc;
+    }
+  }
+  /* return the context */
+  *pctx=ctx;
+  return PAM_SUCCESS;
+}
+
+/* our PAM module configuration */
+struct pld_cfg {
+  int use_first_pass;
+  int try_first_pass;
+  int no_warn;
+  int ignore_unknown_user;
+  int ignore_authinfo_unavail;
+  int debug;
+  uid_t minimum_uid;
+};
+
+static int init(pam_handle_t *pamh,int flags,int argc,const char **argv,
+                struct pld_cfg *cfg,struct pld_ctx **ctx,const char **username,
+                const char **service)
+{
+  int i;
+  int rc;
+  struct passwd *pwent;
+  /* initialise config with defaults */
+  cfg->use_first_pass=0;
+  cfg->try_first_pass=0;
+  cfg->no_warn=0;
+  cfg->ignore_unknown_user=0;
+  cfg->ignore_authinfo_unavail=0;
+  cfg->debug=0;
+  cfg->minimum_uid=0;
+  /* go over arguments */
+  for (i=0;i<argc;i++)
+  {
+    if (strcmp(argv[i],"use_first_pass")==0)
+      cfg->use_first_pass=1;
+    else if (strcmp(argv[i],"try_first_pass")==0)
+      cfg->try_first_pass=1;
+    else if (strcmp(argv[i],"use_authtok")==0)
+      /* ignore, this option is used by pam_get_authtok() internally */;
+    else if (strcmp(argv[i],"no_warn")==0)
+      cfg->no_warn=1;
+    else if (strcmp(argv[i],"ignore_unknown_user")==0)
+      cfg->ignore_unknown_user=1;
+    else if (strcmp(argv[i],"ignore_authinfo_unavail")==0)
+      cfg->ignore_authinfo_unavail=1;
+    else if (strcmp(argv[i],"debug")==0)
+      cfg->debug=1;
+    else if (strncmp(argv[i],"minimum_uid=",12) == 0)
+      cfg->minimum_uid=(uid_t)atoi(argv[i]+12);
+    else
+      pam_syslog(pamh,LOG_ERR,"unknown option: %s",argv[i]);
+  }
+  /* check flags */
+  if (flags&PAM_SILENT)
+    cfg->no_warn=1;
+  /* get user name */
+  rc=pam_get_user(pamh,username,NULL);
+  if (rc!=PAM_SUCCESS)
+  {
+    pam_syslog(pamh,LOG_ERR,"failed to get user name: %s",pam_strerror(pamh,rc));
+    return rc;
+  }
+  if ((*username==NULL)||((*username)[0]=='\0'))
+  {
+    pam_syslog(pamh,LOG_ERR,"got empty user name");
+    return PAM_USER_UNKNOWN;
+  }
+  /* check uid */
+  if (cfg->minimum_uid>0)
+  {
+    pwent=pam_modutil_getpwnam(args->pamh,*username);
+    if ((pwent!=NULL)&&(pwent->pw_uid<cfg->minimum_uid))
+    {
+      if (cfg->debug)
+        pam_syslog(pamh,LOG_DEBUG,"uid below minimum_uid; user=%s uid=%ld",*username,(long)pwent->pw_uid);
+      return cfg->ignore_unknown_user?PAM_IGNORE:PAM_USER_UNKNOWN;
+    }
+  }
+  /* get our context */
+  rc=ctx_get(pamh,*username,ctx);
+  if (rc!=PAM_SUCCESS)
+    return rc;
+  /* get service name */
+  rc=pam_get_item(pamh,PAM_SERVICE,(const void **)service);
+  if (rc!=PAM_SUCCESS)
+  {
+    pam_syslog(pamh,LOG_ERR,"failed to get service name: %s",pam_strerror(pamh,rc));
+    return rc;
+  }
+  return PAM_SUCCESS;
+}
 
 /* map a NSLCD PAM status code to a PAM status code */
 static int nslcd2pam_rc(int rc)
@@ -101,137 +246,14 @@ static int nslcd2pam_rc(int rc)
   }
 }
 
-/* clear the context to all empty values */
-static void ctx_clear(pld_ctx *ctx)
-{
-  if (ctx->user)
-  {
-    free(ctx->user);
-    ctx->user=NULL;
-  }
-  if (ctx->oldpassword)
-  {
-    memset(ctx->oldpassword,0,strlen(ctx->oldpassword));
-    free(ctx->oldpassword);
-    ctx->oldpassword=NULL;
-  }
-  ctx->dn=NULL;
-  ctx->tmpluser=NULL;
-  ctx->authzmsg=NULL;
-  ctx->authok=0;
-  ctx->authz=0;
-}
-
-/* free the context (this is installed as handler into PAM) */
-static void ctx_free(pam_handle_t *UNUSED(pamh),void *data,int UNUSED(err))
-{
-  pld_ctx *ctx=data;
-  ctx_clear(ctx);
-  free(ctx);
-}
-
-/* try to get the module's context, returns a PAM status code */
-static int ctx_get(pam_handle_t *pamh,const char *username,pld_ctx **pctx)
-{
-  pld_ctx *ctx=NULL;
-  int rc;
-  /* try to get the context from PAM */
-  rc=pam_get_data(pamh, PLD_CTX,(const void **)&ctx);
-  if ((rc==PAM_SUCCESS)&&(ctx!=NULL))
-  {
-    /* if the user is different clear the context */
-    if ((ctx->user!=NULL)&&(strcmp(ctx->user,username)!=0))
-      ctx_clear(ctx);
-  }
-  else
-  {
-    /* allocate a new context */
-    ctx=calloc(1,sizeof(*ctx));
-    ctx_clear(ctx);
-    if (!ctx)
-      return PAM_BUF_ERR;
-    /* store the new context with the handler to free it */
-    rc=pam_set_data(pamh,PLD_CTX,ctx,ctx_free);
-    if (rc!=PAM_SUCCESS)
-      ctx_free(pamh,ctx,0);
-  }
-  if (rc==PAM_SUCCESS)
-    *pctx=ctx;
-  return rc;
-}
-
-/* ask the user for an authentication token (password) */
-static int my_pam_get_authtok(pam_handle_t *pamh,int flags,char *prompt1,char *prompt2,const char **pwd)
-{
-  int rc;
-  char *p;
-  struct pam_message msg[1], *pmsg[1];
-  struct pam_response *resp;
-  struct pam_conv *conv;
-
-  *pwd=NULL;
-
-  rc=pam_get_item(pamh,PAM_CONV,(const void **)&conv);
-  if (rc==PAM_SUCCESS) {
-    pmsg[0]=&msg[0];
-    msg[0].msg_style=PAM_PROMPT_ECHO_OFF;
-    msg[0].msg=prompt1;
-    resp=NULL;
-    rc=conv->conv(1,
-       (const struct pam_message **)pmsg,
-       &resp,conv->appdata_ptr);
-  } else {
-    return rc;
-  }
-
-  if (resp!=NULL) {
-    if ((flags & PAM_DISALLOW_NULL_AUTHTOK) && resp[0].resp==NULL)
-    {
-      free(resp);
-      return PAM_AUTH_ERR;
-    }
-
-    p=resp[0].resp;
-    resp[0].resp=NULL;
-    free(resp);
-  } else {
-    return PAM_CONV_ERR;
-  }
-
-  if (prompt2) {
-    msg[0].msg=prompt2;
-    resp=NULL;
-    rc=conv->conv(1,
-       (const struct pam_message **) pmsg,
-       &resp, conv->appdata_ptr);
-    if (resp && resp[0].resp && !strcmp(resp[0].resp, p))
-      rc=PAM_SUCCESS;
-    else
-      rc=PAM_AUTHTOK_RECOVERY_ERR;
-    if (resp) {
-      if (resp[0].resp) {
-        (void) memset(resp[0].resp, 0, strlen(resp[0].resp));
-        free(resp[0].resp);
-      }
-      free(resp);
-    }
-  }
-
-  if (rc==PAM_SUCCESS)
-    *pwd=p;
-  else if (p) {
-    memset(p, 0, strlen(p));
-    free(p);
-  }
-
-  return rc;
-}
-
 /* perform an authentication call over nslcd */
-static int nslcd_request_authc(pld_ctx *ctx,const char *username,
-                               const char *service,const char *passwd)
+static int nslcd_request_authc(pam_handle_t *pamh,struct pld_ctx *ctx,struct pld_cfg *cfg,
+                               const char *username,const char *service,
+                               const char *passwd)
 {
   PAM_REQUEST(NSLCD_ACTION_PAM_AUTHC,
+    /* log debug message */
+    pam_syslog(pamh,LOG_DEBUG,"nslcd authentication; user=%s",username),
     /* write the request parameters */
     WRITE_STRING(fp,username);
     WRITE_STRING(fp,ctx->dn);
@@ -245,124 +267,15 @@ static int nslcd_request_authc(pld_ctx *ctx,const char *username,
     READ_BUF_STRING(fp,ctx->authzmsg);)
 }
 
-int pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,const char **argv)
-{
-  int rc;
-  const char *username,*svc;
-  char *passwd=NULL;
-  int first_pass=0,ignore_flags=0;
-  int i;
-  pld_ctx *ctx;
-  /* go over arguments */
-  for (i=0;i<argc;i++)
-  {
-    if (strcmp(argv[i],"use_first_pass")==0)
-      first_pass|=USE_FIRST;
-    else if (strcmp(argv[i],"try_first_pass")==0)
-      first_pass|=TRY_FIRST;
-    else if (strcmp(argv[i],"ignore_unknown_user")==0)
-      ignore_flags|=IGNORE_UNKNOWN;
-    else if (strcmp(argv[i],"ignore_authinfo_unavail")==0)
-      ignore_flags|=IGNORE_UNAVAIL;
-    else if (strcmp(argv[i], "no_warn")==0)
-      /* ignore */;
-    else if (strcmp(argv[i],"debug")==0)
-      /* ignore */;
-    else
-      syslog(LOG_AUTHPRIV|LOG_ERR,"unknown option: %s",argv[i]);
-  }
-  /* get user name */
-  rc=pam_get_user(pamh,(const char **)&username,NULL);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-  /* get our context */
-  rc=ctx_get(pamh,username,&ctx);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-  /* get service name */
-  rc=pam_get_item(pamh,PAM_SERVICE,(const void **)&svc);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-  /* try twice */
-  for (i=0;i<2;i++)
-  {
-    if (!first_pass)
-    {
-      rc=my_pam_get_authtok(pamh,flags,i==0?"Password: ":"LDAP Password: ",NULL,(const char **)&passwd);
-      if (rc!=PAM_SUCCESS)
-        return rc;
-      /* exit loop after trying this password */
-      i=2;
-      /* store password */
-      pam_set_item(pamh,PAM_AUTHTOK,passwd);
-      /* clear and free password */
-      memset(passwd,0,strlen(passwd));
-      free(passwd);
-    }
-    rc=pam_get_item(pamh,PAM_AUTHTOK,(const void **)&passwd);
-    if (rc==PAM_SUCCESS)
-    {
-      rc=nslcd_request_authc(ctx,username,svc,passwd);
-      if (rc==PAM_SUCCESS)
-        rc=ctx->authok;
-      if ((rc==PAM_AUTHINFO_UNAVAIL)&&(ignore_flags&IGNORE_UNAVAIL))
-        rc=PAM_IGNORE;
-      else if ((rc==PAM_USER_UNKNOWN)&&(ignore_flags&IGNORE_UNKNOWN))
-        rc=PAM_IGNORE;
-    }
-    if ((rc==PAM_SUCCESS)||(first_pass&USE_FIRST))
-      break;
-    first_pass=0;
-  }
-  /* save username */
-  if (rc==PAM_SUCCESS) {
-    ctx->user=strdup(username);
-    /* if password change is required, save old password in context */
-    if (ctx->authz==PAM_NEW_AUTHTOK_REQD)
-      ctx->oldpassword=strdup(passwd);
-  }
-  /* update caller's idea of the user name */
-  if ( (rc==PAM_SUCCESS) && ctx->tmpluser && ctx->tmpluser[0] &&
-       (strcmp(ctx->tmpluser,username)!=0) ) {
-    rc=pam_set_item(pamh,PAM_USER,ctx->tmpluser);
-  }
-  return rc;
-}
-
-/* called to update the authentication credentials */
-int pam_sm_setcred(pam_handle_t UNUSED(*pamh),int UNUSED(flags),
-                   int UNUSED(argc),const char UNUSED(**argv))
-{
-  /* we don't need to do anything here */
-  return PAM_SUCCESS;
-}
-
-static int my_pam_warn(
-  struct pam_conv *aconv, const char *message, int style, int no_warn)
-{
-  struct pam_message msg, *pmsg;
-  struct pam_response *resp;
-
-  if (no_warn)
-    return PAM_SUCCESS;
-
-  pmsg=&msg;
-
-  msg.msg_style=style;
-  msg.msg=(char *)message;
-  resp=NULL;
-
-  return aconv->conv(1,
-          (const struct pam_message **) &pmsg,
-          &resp, aconv->appdata_ptr);
-}
-
 /* perform an authorisation call over nslcd */
-static int nslcd_request_authz(pld_ctx *ctx,const char *username,
-                               const char *service,const char *ruser,
-                               const char *rhost,const char *tty)
+static int nslcd_request_authz(pam_handle_t *pamh,struct pld_ctx *ctx,struct pld_cfg *cfg,
+                               const char *username,const char *service,
+                               const char *ruser,const char *rhost,
+                               const char *tty)
 {
   PAM_REQUEST(NSLCD_ACTION_PAM_AUTHZ,
+    /* log debug message */
+    pam_syslog(pamh,LOG_DEBUG,"nslcd authorisation; user=%s",username),
     /* write the request parameters */
     WRITE_STRING(fp,username);
     WRITE_STRING(fp,ctx->dn);
@@ -377,108 +290,18 @@ static int nslcd_request_authz(pld_ctx *ctx,const char *username,
     READ_BUF_STRING(fp,ctx->authzmsg);)
 }
 
-int pam_sm_acct_mgmt(pam_handle_t *pamh,int flags,int argc,const char **argv)
-{
-  int rc;
-  const char *username,*svc,*ruser,*rhost,*tty;
-  int no_warn=0, ignore_flags=0;
-  int i;
-  struct pam_conv *appconv;
-  pld_ctx *ctx=NULL, ctx2;
-
-  for (i=0;i<argc;i++)
-  {
-    if (strcmp(argv[i],"use_first_pass")==0)
-      ;
-    else if (strcmp(argv[i],"try_first_pass")==0)
-      ;
-    else if (strcmp(argv[i],"no_warn")==0)
-      no_warn=1;
-    else if (strcmp(argv[i],"ignore_unknown_user")==0)
-      ignore_flags|=IGNORE_UNKNOWN;
-    else if (strcmp(argv[i],"ignore_authinfo_unavail")==0)
-      ignore_flags|=IGNORE_UNAVAIL;
-    else if (strcmp(argv[i],"debug")==0)
-      ;
-    else
-      syslog(LOG_AUTHPRIV|LOG_ERR,"unknown option: %s",argv[i]);
-  }
-
-  if (flags&PAM_SILENT)
-    no_warn=1;
-
-  rc=pam_get_item(pamh,PAM_CONV,(const void **)&appconv);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_get_user(pamh,(const char **)&username,NULL);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  if ((username==NULL)||(username[0]=='\0'))
-    return PAM_USER_UNKNOWN;
-
-  rc=ctx_get(pamh,username,&ctx);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_get_item(pamh,PAM_SERVICE,(const void **)&svc);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_get_item (pamh,PAM_RUSER,(const void **)&ruser);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_get_item (pamh,PAM_RHOST,(const void **)&rhost);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_get_item (pamh,PAM_TTY,(const void **)&tty);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  ctx2.dn=ctx->dn;
-  ctx2.user=ctx->user;
-  rc=nslcd_request_authz(&ctx2,username,svc,ruser,rhost,tty);
-  if ((rc==PAM_AUTHINFO_UNAVAIL)&&(ignore_flags&IGNORE_UNAVAIL))
-    rc=PAM_IGNORE;
-  else if ((rc==PAM_USER_UNKNOWN)&&(ignore_flags&IGNORE_UNKNOWN))
-    rc=PAM_IGNORE;
-  if (rc!=PAM_SUCCESS)
-  {
-    if (rc!=PAM_IGNORE)
-      my_pam_warn(appconv,"LDAP authorization failed",PAM_ERROR_MSG,no_warn);
-  }
-  else
-  {
-    rc=ctx2.authz;
-    if (ctx2.authzmsg && ctx2.authzmsg[0])
-      my_pam_warn(appconv,ctx2.authzmsg,PAM_TEXT_INFO,no_warn);
-    if (ctx2.authz==PAM_SUCCESS)
-    {
-      rc=ctx->authz;
-      if (ctx->authzmsg && ctx->authzmsg[0])
-        my_pam_warn(appconv,ctx->authzmsg,PAM_TEXT_INFO,no_warn);
-    }
-  }
-
-  /* update caller's idea of the user name */
-  if ( (rc==PAM_SUCCESS) && ctx->tmpluser && ctx->tmpluser[0] &&
-       (strcmp(ctx->tmpluser,username)!=0) ) {
-    rc=pam_set_item(pamh,PAM_USER,ctx->tmpluser);
-  }
-  return rc;
-}
-
 /* do a session nslcd request (open or close) */
-static int nslcd_request_sess(pld_ctx *ctx,int action,const char *service,
-                              const char *tty, const char *rhost,
+static int nslcd_request_sess(pam_handle_t *pamh,struct pld_ctx *ctx,struct pld_cfg *cfg,int action,
+                              const char *username,const char *service,
+                              const char *tty,const char *rhost,
                               const char *ruser)
 {
   PAM_REQUEST(action,
+    /* log debug message */
+    pam_syslog(pamh,LOG_DEBUG,"nslcd session %s; user=%s",
+          (action==NSLCD_ACTION_PAM_SESS_O)?"open":"clode",username),
     /* write the request parameters */
-    WRITE_STRING(fp,ctx->user);
+    WRITE_STRING(fp,username);
     WRITE_STRING(fp,ctx->dn);
     WRITE_STRING(fp,service);
     WRITE_STRING(fp,tty);
@@ -489,100 +312,14 @@ static int nslcd_request_sess(pld_ctx *ctx,int action,const char *service,
     READ_INT32(fp,ctx->sessid))
 }
 
-static int pam_sm_session(
-  pam_handle_t *pamh, int flags, int argc, const char **argv,
-  int action, int *no_warn)
-{
-  int rc;
-  const char *username;
-  int ignore_flags=0;
-  int i;
-  pld_ctx *ctx=NULL;
-  const char *service=NULL,*tty=NULL,*rhost=NULL,*ruser=NULL;
-
-  for (i=0;i<argc;i++)
-  {
-    if (strcmp(argv[i],"use_first_pass")==0)
-      ;
-    else if (strcmp(argv[i],"try_first_pass")==0)
-      ;
-    else if (strcmp(argv[i],"no_warn")==0)
-      *no_warn=1;
-    else if (strcmp(argv[i],"ignore_unknown_user")==0)
-      ignore_flags|=IGNORE_UNKNOWN;
-    else if (strcmp(argv[i],"ignore_authinfo_unavail")==0)
-      ignore_flags|=IGNORE_UNAVAIL;
-    else if (strcmp(argv[i],"debug")==0)
-      ;
-    else
-      syslog(LOG_AUTHPRIV|LOG_ERR,"unknown option: %s",argv[i]);
-  }
-
-  if (flags & PAM_SILENT)
-    *no_warn=1;
-
-  rc=pam_get_user(pamh,(const char **)&username,NULL);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  if ((username==NULL)||(username[0]=='\0'))
-    return PAM_USER_UNKNOWN;
-
-  rc=ctx_get(pamh,username,&ctx);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-  /* read PAM information */
-  pam_get_item(pamh,PAM_SERVICE,(const void **)&service);
-  pam_get_item(pamh,PAM_TTY,(const void **)&tty);
-  pam_get_item(pamh,PAM_RHOST,(const void **)&rhost);
-  pam_get_item(pamh,PAM_RUSER,(const void **)&ruser);
-  /* do the nslcd request */
-  rc=nslcd_request_sess(ctx,action,service,tty,rhost,ruser);
-  if ((rc==PAM_AUTHINFO_UNAVAIL)&&(ignore_flags&IGNORE_UNAVAIL))
-    rc=PAM_IGNORE;
-  else if ((rc==PAM_USER_UNKNOWN)&&(ignore_flags&IGNORE_UNKNOWN))
-    rc=PAM_IGNORE;
-  return rc;
-}
-
-int pam_sm_open_session(
-  pam_handle_t *pamh, int flags, int argc, const char **argv)
-{
-  int rc, no_warn=0;
-  struct pam_conv *appconv;
-
-  rc=pam_get_item(pamh,PAM_CONV,(const void **)&appconv);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_sm_session(pamh,flags,argc,argv,NSLCD_ACTION_PAM_SESS_O,&no_warn);
-  if ((rc!=PAM_SUCCESS)&&(rc!=PAM_IGNORE))
-    my_pam_warn(appconv,"LDAP open_session failed",PAM_ERROR_MSG,no_warn);
-  return rc;
-}
-
-int pam_sm_close_session(
-  pam_handle_t *pamh, int flags, int argc, const char **argv)
-{
-  int rc, no_warn=0;
-  struct pam_conv *appconv;
-
-  rc=pam_get_item(pamh,PAM_CONV,(const void **)&appconv);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_sm_session(pamh,flags,argc,argv,NSLCD_ACTION_PAM_SESS_C,&no_warn);
-  if ((rc!=PAM_SUCCESS)&&(rc!=PAM_IGNORE))
-    my_pam_warn(appconv,"LDAP close_session failed",PAM_ERROR_MSG,no_warn);
-  return rc;
-}
-
 /* do a password modification nslcd call */
-static int nslcd_request_pwmod(pld_ctx *ctx,const char *username,
-                               const char *service,const char *oldpasswd,
-                               const char *newpasswd)
+static int nslcd_request_pwmod(pam_handle_t *pamh,struct pld_ctx *ctx,struct pld_cfg *cfg,
+                               const char *username,const char *service,
+                               const char *oldpasswd,const char *newpasswd)
 {
   PAM_REQUEST(NSLCD_ACTION_PAM_PWMOD,
+    /* log debug message */
+    pam_syslog(pamh,LOG_DEBUG,"nslcd password modify; user=%s",username),
     /* write the request parameters */
     WRITE_STRING(fp,username);
     WRITE_STRING(fp,ctx->dn);
@@ -596,28 +333,191 @@ static int nslcd_request_pwmod(pld_ctx *ctx,const char *username,
     READ_BUF_STRING(fp,ctx->authzmsg);)
 }
 
-/* ensure that the context includes and oldpassword field */
-static const char *get_old_password(pam_handle_t *pamh, int flags,pld_ctx *ctx)
+/* PAM authentication check */
+int pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,const char **argv)
 {
   int rc;
-  const char *oldpassword;
-  /* if we already have an old password we are done */
-  if ((ctx->oldpassword!=NULL)&&(*ctx->oldpassword!='\0'))
-    return ctx->oldpassword;
-  /* try to get the old password from the PAM stack */
-  rc=pam_get_item(pamh,PAM_OLDAUTHTOK,(const void **)&oldpassword);
-  if ((rc==PAM_SUCCESS)&&(oldpassword!=NULL)&&(*oldpassword!='\0'))
-    return oldpassword;
-  /* otherwise prompt for it */
-  rc=my_pam_get_authtok(pamh,flags,"(current) LDAP Password: ",NULL,
-                     (const char **)&oldpassword);
-  if ((rc==PAM_SUCCESS)&&(oldpassword!=NULL)&&(*oldpassword!='\0'))
+  struct pld_cfg cfg;
+  struct pld_ctx *ctx;
+  const char *username,*service;
+  char *passwd=NULL;
+  int i;
+  /* set up configuration */
+  rc=init(pamh,flags,argc,argv,&cfg,&ctx,&username,&service);
+  if (rc!=PAM_SUCCESS)
+    return rc;
+  /* try twice */
+  for (i=0;i<2;i++)
   {
-    /* save the password */
-    pam_set_item(pamh,PAM_OLDAUTHTOK,oldpassword);
-    return oldpassword;
+    if ((!cfg.try_first_pass)&&(!cfg.use_first_pass))
+    {
+      rc=pam_get_authtok(pamh,PAM_AUTHTOK,(const char **)&passwd,i==0?"Password: ":"LDAP Password: ");
+      if (rc!=PAM_SUCCESS)
+      {
+        pam_syslog(pamh,LOG_ERR,"failed to get password: %s",pam_strerror(pamh,rc));
+        return rc;
+      }
+      /* exit loop after trying this password */
+      i=2;
+    }
+    rc=pam_get_item(pamh,PAM_AUTHTOK,(const void **)&passwd);
+    if (rc!=PAM_SUCCESS)
+      pam_syslog(pamh,LOG_ERR,"failed to get password: %s",pam_strerror(pamh,rc));
+    if (rc==PAM_SUCCESS)
+    {
+      rc=nslcd_request_authc(pamh,ctx,&cfg,username,service,passwd);
+      if (rc==PAM_SUCCESS)
+      {
+        rc=ctx->authok;
+        if (rc!=PAM_SUCCESS)
+          pam_syslog(pamh,LOG_NOTICE,"%s; user=%s",pam_strerror(pamh,rc),username);
+        else if (cfg.debug)
+          pam_syslog(pamh,LOG_DEBUG,"authentication succeeded");
+      }
+      /* remap error code */
+      if ((rc==PAM_AUTHINFO_UNAVAIL)&&cfg.ignore_authinfo_unavail)
+        rc=PAM_IGNORE;
+      else if ((rc==PAM_USER_UNKNOWN)&&cfg.ignore_unknown_user)
+        rc=PAM_IGNORE;
+    }
+    if ((rc==PAM_SUCCESS)||(cfg.use_first_pass))
+      break;
+    cfg.try_first_pass=0;
   }
-  return NULL;
+  /* save username */
+  if (rc==PAM_SUCCESS)
+  {
+    ctx->user=strdup(username);
+    /* if password change is required, save old password in context */
+    if (ctx->authz==PAM_NEW_AUTHTOK_REQD)
+      ctx->oldpassword=strdup(passwd);
+  }
+  /* update caller's idea of the user name */
+  if ( (rc==PAM_SUCCESS) && ctx->tmpluser && ctx->tmpluser[0] &&
+       (strcmp(ctx->tmpluser,username)!=0) )
+  {
+    pam_syslog(pamh,LOG_INFO,"username changed from %s to %s",username,
+               ctx->tmpluser);
+    rc=pam_set_item(pamh,PAM_USER,ctx->tmpluser);
+  }
+  return rc;
+}
+
+/* called to update the authentication credentials */
+int pam_sm_setcred(pam_handle_t UNUSED(*pamh),int UNUSED(flags),
+                   int UNUSED(argc),const char UNUSED(**argv))
+{
+  /* we don't need to do anything here */
+  return PAM_SUCCESS;
+}
+
+/* PAM authorisation check */
+int pam_sm_acct_mgmt(pam_handle_t *pamh,int flags,int argc,const char **argv)
+{
+  int rc;
+  struct pld_cfg cfg;
+  struct pld_ctx *ctx=NULL,ctx2;
+  const char *username,*service;
+  const char *ruser=NULL,*rhost=NULL,*tty=NULL;
+  /* set up configuration */
+  rc=init(pamh,flags,argc,argv,&cfg,&ctx,&username,&service);
+  if (rc!=PAM_SUCCESS)
+    return rc;
+  /* get more PAM information */
+  pam_get_item(pamh,PAM_RUSER,(const void **)&ruser);
+  pam_get_item(pamh,PAM_RHOST,(const void **)&rhost);
+  pam_get_item(pamh,PAM_TTY,(const void **)&tty);
+  /* call the function with a copy of the context to be able to keep the
+     original context */
+  ctx2.dn=ctx->dn;
+  ctx2.user=ctx->user;
+  rc=nslcd_request_authz(pamh,&ctx2,&cfg,username,service,ruser,rhost,tty);
+  /* remap error code */
+  if ((rc==PAM_AUTHINFO_UNAVAIL)&&cfg.ignore_authinfo_unavail)
+    rc=PAM_IGNORE;
+  else if ((rc==PAM_USER_UNKNOWN)&&cfg.ignore_unknown_user)
+    rc=PAM_IGNORE;
+  if (rc!=PAM_SUCCESS)
+  {
+    if (rc!=PAM_IGNORE)
+      if (!cfg.no_warn)
+        pam_error(pamh,"LDAP authorization failed");
+    return rc;
+  }
+  /* check the returned authorisation value */
+  if (ctx2.authz!=PAM_SUCCESS)
+  {
+    pam_syslog(pamh,LOG_NOTICE,"%s; user=%s",ctx2.authzmsg,username);
+    if (!cfg.no_warn)
+      pam_error(pamh,"%s",ctx2.authzmsg);
+    return ctx2.authz;
+  }
+  /* check the original authorisation check from authentication */
+  if (ctx->authz!=PAM_SUCCESS)
+  {
+    pam_syslog(pamh,LOG_NOTICE,"%s; user=%s",ctx->authzmsg,username);
+    if (!cfg.no_warn)
+      pam_error(pamh,"%s",ctx->authzmsg);
+    return ctx->authz;
+  }
+  if (cfg.debug)
+    pam_syslog(pamh,LOG_DEBUG,"authorization succeeded");
+  /* present any informational messages to the user */
+  if ((ctx2.authzmsg!=NULL)&&(ctx2.authzmsg[0]!='\0')&&(!cfg.no_warn))
+    pam_info(pamh,"%s",ctx2.authzmsg);
+  if ((ctx->authzmsg!=NULL)&&(ctx->authzmsg[0]!='\0')&&(!cfg.no_warn))
+    pam_info(pamh,"%s",ctx->authzmsg);
+  return PAM_SUCCESS;
+}
+
+/* PAM session open/close calls */
+static int pam_sm_session(pam_handle_t *pamh,int flags,int argc,
+                          const char **argv,int action)
+{
+  int rc;
+  struct pld_cfg cfg;
+  struct pld_ctx *ctx;
+  const char *username,*service;
+  const char *tty=NULL,*rhost=NULL,*ruser=NULL;
+  /* set up configuration */
+  rc=init(pamh,flags,argc,argv,&cfg,&ctx,&username,&service);
+  if (rc!=PAM_SUCCESS)
+    return rc;
+  /* get more PAM information */
+  pam_get_item(pamh,PAM_TTY,(const void **)&tty);
+  pam_get_item(pamh,PAM_RHOST,(const void **)&rhost);
+  pam_get_item(pamh,PAM_RUSER,(const void **)&ruser);
+  /* do the nslcd request */
+  rc=nslcd_request_sess(pamh,ctx,&cfg,action,username,service,tty,rhost,ruser);
+  if (rc!=PAM_SUCCESS)
+    pam_syslog(pamh,LOG_NOTICE,"%s; user=%s",pam_strerror(pamh,rc),username);
+  else if (cfg.debug)
+    pam_syslog(pamh,LOG_DEBUG,"session %s succeeded; session_id=%d",
+               (action==NSLCD_ACTION_PAM_SESS_O)?"open":"clode",ctx->sessid);
+  /* remap error code */
+  if ((rc==PAM_AUTHINFO_UNAVAIL)&&cfg.ignore_authinfo_unavail)
+    rc=PAM_IGNORE;
+  else if ((rc==PAM_USER_UNKNOWN)&&cfg.ignore_unknown_user)
+    rc=PAM_IGNORE;
+  if ((rc!=PAM_SUCCESS)&&(rc!=PAM_IGNORE))
+    if (!cfg.no_warn)
+      pam_error(pamh,"LDAP %s session failed",
+                     (action==NSLCD_ACTION_PAM_SESS_O)?"open":"clode");
+  return rc;
+}
+
+/* PAM session open call */
+int pam_sm_open_session(
+  pam_handle_t *pamh,int flags,int argc,const char **argv)
+{
+  return pam_sm_session(pamh,flags,argc,argv,NSLCD_ACTION_PAM_SESS_O);
+}
+
+/* PAM session close call */
+int pam_sm_close_session(
+  pam_handle_t *pamh,int flags,int argc,const char **argv)
+{
+  return pam_sm_session(pamh,flags,argc,argv,NSLCD_ACTION_PAM_SESS_C);
 }
 
 /* Change the password of the user. This function is first called with
@@ -628,73 +528,48 @@ static const char *get_old_password(pam_handle_t *pamh, int flags,pld_ctx *ctx)
 int pam_sm_chauthtok(pam_handle_t *pamh,int flags,int argc,const char **argv)
 {
   int rc;
+  struct pld_cfg cfg;
+  struct pld_ctx *ctx;
   const char *username,*service;
-  const char *oldpassword=NULL;
-  const char *newpassword=NULL;
-  int first_pass=0, no_warn=0, ignore_flags=0;
-  int i;
-  struct pam_conv *appconv;
-  pld_ctx *ctx=NULL;
-
-  for (i=0;i<argc;i++)
-  {
-    if (strcmp(argv[i],"use_first_pass")==0)
-      first_pass|=USE_FIRST;
-    else if (strcmp(argv[i],"try_first_pass")==0)
-      first_pass|=TRY_FIRST;
-    else if (strcmp(argv[i],"use_authtok")==0)
-      first_pass|=USE_TOKEN;
-    else if (strcmp(argv[i],"no_warn")==0)
-      no_warn=1;
-    else if (strcmp(argv[i],"ignore_unknown_user")==0)
-      ignore_flags|=IGNORE_UNKNOWN;
-    else if (strcmp(argv[i],"ignore_authinfo_unavail")==0)
-      ignore_flags|=IGNORE_UNAVAIL;
-    else if (strcmp(argv[i],"debug")==0)
-      ;
-    else
-      syslog(LOG_AUTHPRIV|LOG_ERR,"unknown option: %s",argv[i]);
-  }
-
-  if (flags&PAM_SILENT)
-    no_warn=1;
-
-  rc=pam_get_item(pamh,PAM_CONV,(const void **)&appconv);
+  const char *oldpassword=NULL,*newpassword=NULL;
+  struct passwd *pwent;
+  /* set up configuration */
+  rc=init(pamh,flags,argc,argv,&cfg,&ctx,&username,&service);
   if (rc!=PAM_SUCCESS)
     return rc;
-
-  rc=pam_get_user(pamh,(const char **)&username,NULL);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  if (username==NULL)
-    return PAM_USER_UNKNOWN;
-
-  rc=ctx_get(pamh,username,&ctx);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-
-  rc=pam_get_item(pamh,PAM_SERVICE,(const void **)&service);
-  if (rc!=PAM_SUCCESS)
-    return rc;
-  /* TODO: if we are root we may want to authenticate with the LDAP
-           administrator password (this shouldn't be a problem because
-           root is unlikely to be in LDAP anyway but perhaps we can check
-           the requested username and only use the administrator if that
-           isn't root) */
   /* prelimenary check, just see if we can connect to the LDAP server
      and authenticate with the current password */
   if (flags&PAM_PRELIM_CHECK)
   {
-    /* get old (current) password */
-    oldpassword=get_old_password(pamh,flags,ctx);
-    /* check the old password */
-    rc=nslcd_request_authc(ctx,username,service,oldpassword);
+    /* see if the user is trying to modify another user's password */
+    pwent=getpwnam(username);
+    if ((pwent!=NULL)&&(pwent->pw_uid!=getuid()))
+    {
+      /* try to  authenticate with the LDAP administrator password by passing
+         an empty username to the authc request */
+      rc=pam_get_authtok(pamh,PAM_OLDAUTHTOK,&oldpassword,"LDAP administrator password: ");
+      if (rc!=PAM_SUCCESS)
+        return rc;
+      username="";
+    }
+    else if ((ctx->oldpassword!=NULL)&&(*ctx->oldpassword!='\0'))
+      /* we already have an old password stored (from a previous
+         authentication phase) so we'll use that */
+      oldpassword=ctx->oldpassword;
+    else
+    {
+      rc=pam_get_authtok(pamh,PAM_OLDAUTHTOK,(const char **)&oldpassword,"(current) LDAP Password: ");
+      if (rc!=PAM_SUCCESS)
+        return rc;
+    }
+    /* try authenticating */
+    rc=nslcd_request_authc(pamh,ctx,&cfg,username,service,oldpassword);
     if (rc==PAM_SUCCESS)
       rc=ctx->authok;
-    if ((rc==PAM_AUTHINFO_UNAVAIL)&&(ignore_flags&IGNORE_UNAVAIL))
+    /* remap error code */
+    if ((rc==PAM_AUTHINFO_UNAVAIL)&&cfg.ignore_authinfo_unavail)
       rc=PAM_IGNORE;
-    else if ((rc==PAM_USER_UNKNOWN)&&(ignore_flags&IGNORE_UNKNOWN))
+    else if ((rc==PAM_USER_UNKNOWN)&&cfg.ignore_unknown_user)
       rc=PAM_IGNORE;
     /* TODO: figure out when to return PAM_TRY_AGAIN */
     /* TODO: if password is incorrect (NSLCD_PAM_AUTH_ERR) log that */
@@ -709,18 +584,26 @@ int pam_sm_chauthtok(pam_handle_t *pamh,int flags,int argc,const char **argv)
   if (rc!=PAM_SUCCESS)
     return rc;
   /* perform the password modification */
-  rc=nslcd_request_pwmod(ctx,username,service,oldpassword,newpassword);
+  rc=nslcd_request_pwmod(pamh,ctx,&cfg,username,service,oldpassword,newpassword);
   if (rc==PAM_SUCCESS)
     rc=ctx->authz;
   else
     ctx->authzmsg=(char *)pam_strerror(pamh,rc);
-  if ((rc==PAM_AUTHINFO_UNAVAIL)&&(ignore_flags&IGNORE_UNAVAIL))
+  /* remap error code */
+  if ((rc==PAM_AUTHINFO_UNAVAIL)&&cfg.ignore_authinfo_unavail)
     rc=PAM_IGNORE;
-  else if ((rc==PAM_USER_UNKNOWN)&&(ignore_flags&IGNORE_UNKNOWN))
+  else if ((rc==PAM_USER_UNKNOWN)&&cfg.ignore_unknown_user)
     rc=PAM_IGNORE;
-  else
-    my_pam_warn(appconv,ctx->authzmsg,PAM_ERROR_MSG,no_warn);
-  return rc;
+  /* check the returned value */
+  if (rc!=PAM_SUCCESS)
+  {
+    pam_syslog(pamh,LOG_NOTICE,"password change failed: %s; user=%s",ctx->authzmsg,username);
+    if ((rc!=PAM_IGNORE)&&(!cfg.no_warn))
+      pam_error(pamh,"%s",ctx->authzmsg);
+    return rc;
+  }
+  pam_syslog(pamh,LOG_NOTICE,"password changed for %s",username);
+  return PAM_SUCCESS;
 }
 
 #ifdef PAM_STATIC
